#!/usr/bin/env python
# Copyright 2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA metastatus.
"""
from pyramid.view import view_config

from paasta_tools.paasta_metastatus import get_output


@view_config(route_name='metastatus', request_method='GET', renderer='json')
def metastatus(request):
    cmd_args = request.swagger_data.get('cmd_args', None)
    output, exit_code = get_output(cmd_args)
    return {
        "output": output,
        "exit_code": exit_code,
    }



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Dict

import aiohttp
from mypy_extensions import TypedDict

from paasta_tools.utils import get_user_agent

HACHECK_CONN_TIMEOUT = 30
HACHECK_READ_TIMEOUT = 10

SpoolInfo = TypedDict(
    'SpoolInfo',
    {
        'service': str,
        'state': str,
        'since': float,
        'until': float,
        'reason': str,
    },
    total=False,
)


async def post_spool(url: str, status: str, data: Dict['str', 'str']) -> None:
    async with aiohttp.ClientSession(
        conn_timeout=HACHECK_CONN_TIMEOUT,
        read_timeout=HACHECK_READ_TIMEOUT,
    ) as session:
        async with session.post(
            url,
            data=data,
            headers={'User-Agent': get_user_agent()},
        ) as resp:
            resp.raise_for_status()


async def get_spool(spool_url: str) -> SpoolInfo:
    """Query hacheck for the state of a task, and parse the result into a dictionary."""
    if spool_url is None:
        return None

    # TODO: aiohttp says not to create a session per request. Fix this.
    async with aiohttp.ClientSession(
        conn_timeout=HACHECK_CONN_TIMEOUT,
        read_timeout=HACHECK_READ_TIMEOUT,
    ) as session:
        async with session.get(
            spool_url,
            headers={'User-Agent': get_user_agent()},
        ) as response:
            if response.status == 200:
                return {
                    'state': 'up',
                }

            regex = ''.join([
                "^",
                r"Service (?P<service>.+)",
                r" in (?P<state>.+) state",
                r"(?: since (?P<since>[0-9.]+))?",
                r"(?: until (?P<until>[0-9.]+))?",
                r"(?:: (?P<reason>.*))?",
                "$",
            ])

            response_text = await response.text()
            match = re.match(regex, response_text)
            groupdict = match.groupdict()
            info: SpoolInfo = {}
            info['service'] = groupdict['service']
            info['state'] = groupdict['state']
            if 'since' in groupdict:
                info['since'] = float(groupdict['since'] or 0)
            if 'until' in groupdict:
                info['until'] = float(groupdict['until'] or 0)
            if 'reason' in groupdict:
                info['reason'] = groupdict['reason']
            return info



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import logging
import os
import sys
from typing import Mapping
from typing import Sequence

from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import create_kubernetes_secret_signature
from paasta_tools.kubernetes_tools import create_secret
from paasta_tools.kubernetes_tools import get_kubernetes_secret_signature
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import update_kubernetes_secret_signature
from paasta_tools.kubernetes_tools import update_secret
from paasta_tools.secret_tools import get_secret_provider
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config


log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Sync paasta secrets into k8s')
    parser.add_argument(
        'service_list', nargs='+',
        help="The list of services to sync secrets for",
        metavar="SERVICE",
    )
    parser.add_argument(
        '-c', '--cluster', dest="cluster", metavar="CLUSTER",
        default=None,
        help="Kubernetes cluster name",
    )
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        dest="verbose", default=False,
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    system_paasta_config = load_system_paasta_config()
    if args.cluster:
        cluster = args.cluster
    else:
        cluster = system_paasta_config.get_cluster()
    secret_provider_name = system_paasta_config.get_secret_provider_name()
    vault_cluster_config = system_paasta_config.get_vault_cluster_config()
    kube_client = KubeClient()
    sys.exit(0) if sync_all_secrets(
        kube_client=kube_client,
        cluster=cluster,
        service_list=args.service_list,
        secret_provider_name=secret_provider_name,
        vault_cluster_config=vault_cluster_config,
        soa_dir=args.soa_dir,
    ) else sys.exit(1)


def sync_all_secrets(
    kube_client: KubeClient,
    cluster: str,
    service_list: Sequence[str],
    secret_provider_name: str,
    vault_cluster_config: Mapping[str, str],
    soa_dir: str,
) -> bool:
    results = []
    for service in service_list:
        results.append(sync_secrets(
            kube_client=kube_client,
            cluster=cluster,
            service=service,
            secret_provider_name=secret_provider_name,
            vault_cluster_config=vault_cluster_config,
            soa_dir=soa_dir,
        ))
    return all(results)


def sync_secrets(
    kube_client: KubeClient,
    cluster: str,
    service: str,
    secret_provider_name: str,
    vault_cluster_config: Mapping[str, str],
    soa_dir: str,
) -> bool:
    secret_dir = os.path.join(soa_dir, service, "secrets")
    secret_provider_kwargs = {
        'vault_cluster_config': vault_cluster_config,
        # TODO: make vault-tools support k8s auth method so we don't have to
        # mount a token in.
        'vault_auth_method': 'token',
        'vault_token_file': '/root/.vault_token',
    }
    secret_provider = get_secret_provider(
        secret_provider_name=secret_provider_name,
        soa_dir=soa_dir,
        service_name=service,
        cluster_names=[cluster],
        secret_provider_kwargs=secret_provider_kwargs,
    )
    if not os.path.isdir(secret_dir):
        log.debug(f"No secrets dir for {service}")
        return True
    with os.scandir(secret_dir) as secret_file_paths:
        for secret_file_path in secret_file_paths:
            if secret_file_path.path.endswith('json'):
                secret = secret_file_path.name.replace('.json', '')
                with open(secret_file_path, 'r') as secret_file:
                    secret_data = json.load(secret_file)
                secret_signature = secret_provider.get_secret_signature_from_data(secret_data)
                if secret_signature:
                    kubernetes_secret_signature = get_kubernetes_secret_signature(
                        kube_client=kube_client,
                        secret=secret,
                        service=service,
                    )
                    if not kubernetes_secret_signature:
                        log.info(f"{secret} for {service} not found, creating")
                        try:
                            create_secret(
                                kube_client=kube_client,
                                secret=secret,
                                service=service,
                                secret_provider=secret_provider,
                            )
                        except ApiException as e:
                            if e.status == 409:
                                log.warning(f"Secret {secret} for {service} already exists")
                            else:
                                raise
                        create_kubernetes_secret_signature(
                            kube_client=kube_client,
                            secret=secret,
                            service=service,
                            secret_signature=secret_signature,
                        )
                    elif secret_signature != kubernetes_secret_signature:
                        log.info(f"{secret} for {service} needs updating as signature changed")
                        update_secret(
                            kube_client=kube_client,
                            secret=secret,
                            service=service,
                            secret_provider=secret_provider,
                        )
                        update_kubernetes_secret_signature(
                            kube_client=kube_client,
                            secret=secret,
                            service=service,
                            secret_signature=secret_signature,
                        )
                    else:
                        log.info(f"{secret} for {service} up to date")
    return True


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
paasta_oom_logger is supposed to be used as a syslog-ng destination.
It looks for OOM events in the log, adds PaaSTA service and instance names
and send JSON-encoded messages the Scribe stream 'tmp_paasta_oom_events'.

syslog-ng.conf:

destination paasta_oom_logger {
  program("exec /usr/bin/paasta_oom_logger" template("${UNIXTIME} ${HOST} ${MESSAGE}\n") );
};

filter f_cgroup_oom {
  match(" killed as a result of limit of ") or match(" invoked oom-killer: ");
};

log {
  source(s_all);
  filter(f_cgroup_oom);
  destination(paasta_oom_logger);
};
"""
import re
import sys
from collections import namedtuple

from clog.loggers import ScribeLogger
from docker.errors import APIError

from paasta_tools.utils import _log
from paasta_tools.utils import DEFAULT_LOGLEVEL
from paasta_tools.utils import get_docker_client
from paasta_tools.utils import load_system_paasta_config


try:
    import yelp_meteorite
except ImportError:
    # Sorry to any non-yelpers but you won't
    # get metrics emitted as our metrics lib
    # is currently not open source
    yelp_meteorite = None


LogLine = namedtuple(
    'LogLine', [
        'timestamp', 'hostname', 'container_id',
        'cluster', 'service', 'instance', 'process_name',
    ],
)


def capture_oom_events_from_stdin():
    process_name_regex = re.compile(r'^\d+\s[a-zA-Z0-9\-]+\s.*\]\s(.+)\sinvoked\soom-killer:')
    oom_regex = re.compile(r'^(\d+)\s([a-zA-Z0-9\-]+)\s.*Task in /docker/(\w{12})\w+ killed as a')
    process_name = ''

    while True:
        syslog = sys.stdin.readline()
        if not syslog:
            break
        r = process_name_regex.search(syslog)
        if r:
            process_name = r.group(1)
        r = oom_regex.search(syslog)
        if r:
            yield (int(r.group(1)), r.group(2), r.group(3), process_name)
            process_name = ''


def get_container_env_as_dict(docker_inspect):
    env_vars = {}
    config = docker_inspect.get('Config')
    if config is not None:
        env = config.get('Env', [])
        for i in env:
            name, _, value = i.partition('=')
            env_vars[name] = value
    return env_vars


def log_to_scribe(logger, log_line):
    """Send the event to 'tmp_paasta_oom_events'."""
    line = ('{"timestamp": %d, "hostname": "%s", "container_id": "%s", "cluster": "%s", '
            '"service": "%s", "instance": "%s", "process_name": "%s"}' % (
                log_line.timestamp, log_line.hostname, log_line.container_id,
                log_line.cluster, log_line.service, log_line.instance,
                log_line.process_name,
            ))
    logger.log_line('tmp_paasta_oom_events', line)


def log_to_paasta(log_line):
    """Add the event to the standard PaaSTA logging backend."""
    line = ('oom-killer killed %s on %s (container_id: %s).'
            % (
                'a %s process' % log_line.process_name if log_line.process_name else 'a process',
                log_line.hostname, log_line.container_id,
            ))
    _log(
        service=log_line.service, instance=log_line.instance, component='oom',
        cluster=log_line.cluster, level=DEFAULT_LOGLEVEL, line=line,
    )


def send_sfx_event(service, instance, cluster):
    if yelp_meteorite:
        yelp_meteorite.events.emit_event(
            'paasta.service.oom_events',
            dimensions={
                'paasta_cluster': cluster,
                'paasta_instance': instance,
                'paasta_service': service,
            },
        )


def main():
    scribe_logger = ScribeLogger(host='169.254.255.254', port=1463, retry_interval=5)
    cluster = load_system_paasta_config().get_cluster()
    client = get_docker_client()
    for timestamp, hostname, container_id, process_name in capture_oom_events_from_stdin():
        try:
            docker_inspect = client.inspect_container(resource_id=container_id)
        except (APIError):
            continue
        env_vars = get_container_env_as_dict(docker_inspect)
        service = env_vars.get('PAASTA_SERVICE', 'unknown')
        instance = env_vars.get('PAASTA_INSTANCE', 'unknown')
        log_line = LogLine(
            timestamp=timestamp,
            hostname=hostname,
            container_id=container_id,
            cluster=cluster,
            service=service,
            instance=instance,
            process_name=process_name,
        )
        log_to_scribe(scribe_logger, log_line)
        log_to_paasta(log_line)
        send_sfx_event(service, instance, cluster)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import time

import a_sync
import simplejson as json

from paasta_tools import mesos_tools
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def get_container_info_from_task(task):
    for status in task['statuses']:
        if status['state'] != 'TASK_RUNNING':
            continue
        container_id = status.get('container_status', {}).get('container_id', {}).get('value')
        time_start = status.get('timestamp')
        return container_id, time_start
    return None, None


def get_paasta_service_instance_from_task(task):
    try:
        docker_params = task['container'].get('docker', {}).get('parameters', [])
    except KeyError:
        return None, None
    service, instance = None, None
    for param in docker_params:
        if param['key'] == 'label':
            label = param['value']
            if label.startswith('paasta_service='):
                service = label.split('=')[1]
            if label.startswith('paasta_instance='):
                instance = label.split('=')[1]
    return service, instance


async def get_pool_from_task(task):
    attributes = (await task.slave())['attributes']
    return attributes.get('pool', 'default')


@a_sync.to_blocking
async def get_task_allocation_info():
    tasks = await mesos_tools.get_cached_list_of_running_tasks_from_frameworks()
    info_list = []
    for task in tasks:
        info = {'resources': task['resources']}
        info['mesos_container_id'], info['start_time'] = get_container_info_from_task(task)
        info['paasta_service'], info['paasta_instance'] = get_paasta_service_instance_from_task(task)
        info['paasta_pool'] = await get_pool_from_task(task)
        info_list.append(info)
    return info_list


def main():
    cluster = load_system_paasta_config().get_cluster()
    info_list = get_task_allocation_info()
    timestamp = time.time()
    for info in info_list:
        info['cluster'] = cluster
        info['timestamp'] = timestamp
        paasta_print(json.dumps(info))


if __name__ == '__main__':
    main()



import abc
import functools
import os
import textwrap
import threading
import time
import traceback
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple

import pytimeparse
try:
    from slo_utils.yelpsoa_configs import get_slo_files_from_soaconfigs
    from slo_transcoder.composite_sinks.signalform_detectors import make_signalform_detector_composite_sink
    from slo_transcoder.sources.yelpsoaconfigs import YelpSoaConfigsSource
    SLO_TRANSCODER_LOADED = True
except ImportError:
    SLO_TRANSCODER_LOADED = False

import tempfile

from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.automatic_rollbacks.signalfx import tail_signalfx
from paasta_tools.automatic_rollbacks.slack import SlackDeploymentProcess


def get_slos_for_service(service, soa_dir=DEFAULT_SOA_DIR) -> Generator:
    if not SLO_TRANSCODER_LOADED:
        return

    slo_files = [f for f in get_slo_files_from_soaconfigs(soa_dir) if f.startswith(os.path.join(soa_dir, service))]
    with tempfile.TemporaryDirectory() as temp_dir:
        SFDCS = make_signalform_detector_composite_sink(output_directory=temp_dir)
        sources = [YelpSoaConfigsSource(file) for file in slo_files]
        if not sources:
            return   # SFDCS instantiation crashes with IndexError if sources is an empty list.
        composite_sink = SFDCS(service=service, sources=sources)

        for sink in composite_sink.sinks():
            signalflow, rules = sink.generate_signalflow_signals_and_rules()
            query = textwrap.dedent('\n'.join(signalflow))
            yield sink, query


class SLODemultiplexer:
    def __init__(
        self,
        sink: Any,
        individual_slo_callback: Callable[['SLOWatcher'], None],
        start_timestamp: Optional[float] = None,
        max_duration: float = 3600,
    ) -> None:
        self.sink = sink
        self.individual_slo_callback = individual_slo_callback
        self.max_duration = max_duration
        if start_timestamp is None:
            self.start_timestamp = time.time()
        else:
            self.start_timestamp = start_timestamp

        self.slo_watchers_by_label: Dict[str, 'SLOWatcher'] = {}
        for slo in sink.slos:
            label = sink._get_detector_label(slo)
            watcher = SLOWatcher(slo, individual_slo_callback, self.start_timestamp, label, max_duration=max_duration)
            self.slo_watchers_by_label[label] = watcher

    def process_datapoint(self, props, datapoint, timestamp) -> None:
        slo_label = props['dimensions']['sf_metric'].rsplit('.', 1)[0]
        watcher = self.slo_watchers_by_label[slo_label]
        watcher.process_datapoint(props, datapoint, timestamp)


class SLOWatcher:
    def __init__(
        self,
        slo: Any,
        callback: Callable[['SLOWatcher'], Any],
        start_timestamp: float,
        label: str,
        max_duration: float,
    ) -> None:
        self.slo = slo
        self.window: List[Tuple[float, float]] = []
        self.max_duration = max_duration
        self.failing: Optional[bool] = None
        self.bad_after_mark: Optional[bool] = None
        self.bad_before_mark: Optional[bool] = None
        self.callback = callback
        self.start_timestamp = start_timestamp
        self.label = label

    def process_datapoint(self, props, datapoint, timestamp) -> None:
        self.window.append((timestamp, datapoint))
        self.trim_window()

        if timestamp > self.start_timestamp:
            self.bad_after_mark = self.is_window_bad()
        else:
            self.bad_before_mark = self.is_window_bad()

        old_failing = self.failing
        self.failing = self.bad_after_mark and not self.bad_before_mark

        if self.failing == (not old_failing):
            self.callback(self)

    def trim_window(self) -> None:
        old_window = self.window
        min_ts = max(ts for ts, d in old_window) - self.window_duration()
        new_window = [(ts, d) for (ts, d) in old_window if ts > min_ts]
        self.window = new_window

    def window_duration(self) -> float:
        """Many of our SLOs are defined with durations of 1 hour or more; this is great if you're trying to avoid being
        paged, but not helpful for a deployment that you expect to finish within a few minutes. self.max_duration allows
        us to cap the length of time that we consider. This should make us a bit more sensitive."""
        return min(self.max_duration, pytimeparse.parse(str(self.slo.config.duration)))

    def is_window_bad(self) -> bool:
        bad_datapoints = len([1 for ts, d in self.window if d > self.slo.config.threshold])
        return (bad_datapoints / len(self.window)) >= 1. - self.slo.config.percent_of_duration / 100.


def print_exceptions_wrapper(fn):
    def inner(*args, **kwargs):
        try:
            fn(*args, **kwargs)
        except Exception:
            traceback.print_exc()
            raise
    return inner


def watch_slos_for_service(
    service: str,
    individual_slo_callback: Callable[[str, bool], Any],
    all_slos_callback: Callable[[bool], Any],
    sfx_api_token: str,
    start_timestamp: Optional[float] = None,
) -> Tuple[List[threading.Thread], List[SLOWatcher]]:
    threads = []
    watchers: List[SLOWatcher] = []

    failing = False

    def callback_wrapper(watcher: 'SLOWatcher') -> None:
        nonlocal failing
        old_failing = failing
        new_failing = any(w.failing for w in watchers)
        individual_slo_callback(watcher.label, watcher.failing)

        failing = new_failing

        if new_failing == (not old_failing):
            all_slos_callback(new_failing)

    for sink, query in get_slos_for_service(service):
        demux = SLODemultiplexer(
            sink,
            individual_slo_callback=callback_wrapper,
            start_timestamp=start_timestamp,
        )
        thread = threading.Thread(
            target=print_exceptions_wrapper(functools.partial(
                tail_signalfx,
                query,
                lookback_seconds=30,
                callback=print_exceptions_wrapper(demux.process_datapoint),
                sfx_api_token=sfx_api_token,
            )),
            daemon=True,
        )
        threads.append(thread)
        thread.start()
        watchers.extend(demux.slo_watchers_by_label.values())

    return threads, watchers


class SLOSlackDeploymentProcess(SlackDeploymentProcess, abc.ABC):
    auto_rollback_delay: float

    def get_extra_blocks_for_deployment(self):
        blocks = []
        slo_text = self.get_slo_text(summary=False)
        if slo_text:
            blocks.append({
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": slo_text,
                },
            })
        return blocks

    def get_extra_summary_parts_for_deployment(self) -> List[str]:
        parts = super().get_extra_summary_parts_for_deployment()
        slo_text = self.get_slo_text(summary=True)
        if slo_text:
            parts.append(slo_text)

        return parts

    def get_slo_text(self, summary: bool) -> str:
        slo_watchers = getattr(self, 'slo_watchers', None)
        if slo_watchers is not None and len(slo_watchers) > 0:
            num_failing = len([w for w in slo_watchers if w.failing])

            # Wrap emojis in this subclass so we can select only the emojis or only the detail sections.
            class Emoji(str):
                pass

            if num_failing > 0:
                slo_text_components = [
                    Emoji(":alert:"),
                    f"{num_failing} of {len(slo_watchers)} SLOs are failing",
                ]
            else:

                num_unknown = len([w for w in slo_watchers if w.bad_before_mark is None or w.bad_after_mark is None])
                num_bad_before_mark = len([w for w in slo_watchers if w.bad_before_mark])
                slo_text_components = []
                if num_unknown > 0:
                    slo_text_components.extend([Emoji(":thinking_face:"), f"{num_unknown} SLOs are missing data. "])
                if num_bad_before_mark > 0:
                    slo_text_components.extend([
                        Emoji(":grimacing:"),
                        f"{num_bad_before_mark} SLOs were failing before deploy, and will be ignored.",
                    ])

                remaining = len(slo_watchers) - num_unknown - num_bad_before_mark

                if remaining == len(slo_watchers):
                    slo_text_components = [
                        Emoji(":ok_hand:"),
                        f"All {len(slo_watchers)} SLOs are currently passing.",
                    ]
                else:
                    if remaining > 0:
                        slo_text_components.append(f"The remaining {remaining} SLOs are currently passing.")

            if summary:
                # For summary, only display emojis.
                if self.is_terminal_state(self.state):
                    return ''
                else:
                    return ' '.join([c for c in slo_text_components if isinstance(c, Emoji)])
            else:
                # Display all text for non-summary mode, but hide Emojis if we're in a terminal state, to prevent
                # things like :alert: from blinking until the end of time.
                if self.is_terminal_state(self.state):
                    return ' '.join([c for c in slo_text_components if not isinstance(c, Emoji)])
                else:
                    return ' '.join(slo_text_components)
        else:
            return ""

    def start_slo_watcher_threads(self, service: str) -> None:
        _, self.slo_watchers = watch_slos_for_service(
            service=service,
            individual_slo_callback=self.individual_slo_callback,
            all_slos_callback=self.all_slos_callback,
            sfx_api_token=self.get_signalfx_api_token(),
        )

    @abc.abstractmethod
    def get_signalfx_api_token(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def auto_rollbacks_enabled(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_auto_rollback_delay(self) -> float:
        raise NotImplementedError()

    def any_slo_failing(self) -> bool:
        return self.auto_rollbacks_enabled() and any(w.failing for w in self.slo_watchers)

    def individual_slo_callback(self, label: str, bad: bool) -> None:
        if bad:
            message = f"SLO started failing: {label}"
        else:
            message = f"SLO is now OK: {label}"
        self.notify_users(message)

    def all_slos_callback(self, bad: bool) -> None:
        if bad:
            self.trigger('slos_started_failing')
        else:
            self.trigger('slos_stopped_failing')
        self.update_slack()

    def start_auto_rollback_countdown(self) -> None:
        self.start_timer(self.get_auto_rollback_delay(), 'rollback_slo_failure', "automatically roll back")

    def cancel_auto_rollback_countdown(self) -> None:
        self.cancel_timer('rollback_slo_failure')



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./cleanup_kubernetes_cr.py [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -v, --verbose: Verbose output
"""
import argparse
import logging
import sys
from typing import Sequence

from paasta_tools.kubernetes_tools import CustomResourceDefinition
from paasta_tools.kubernetes_tools import delete_custom_resource
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import list_custom_resources
from paasta_tools.kubernetes_tools import load_custom_resource_definitions
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_all_configs
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Cleanup custom_resources.')
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        dest="verbose", default=False,
    )
    parser.add_argument(
        '-c', '--cluster', default=None,
        help="Cluster to cleanup CRs for",
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    kube_client = KubeClient()

    system_paasta_config = load_system_paasta_config()
    cluster = args.cluster or system_paasta_config.get_cluster()
    custom_resource_definitions = load_custom_resource_definitions(system_paasta_config)
    cleanup_kube_succeeded = cleanup_all_custom_resources(
        kube_client=kube_client,
        soa_dir=soa_dir,
        cluster=cluster,
        custom_resource_definitions=custom_resource_definitions,
    )
    sys.exit(0 if cleanup_kube_succeeded else 1)


def cleanup_all_custom_resources(
    kube_client: KubeClient,
    soa_dir: str,
    cluster: str,
    custom_resource_definitions: Sequence[CustomResourceDefinition],
) -> bool:
    cluster_crds = {
        crd.spec.names.kind
        for crd in
        kube_client.apiextensions.list_custom_resource_definition(
            label_selector="yelp.com/paasta_service",
        ).items
    }
    log.debug(f"CRDs found: {cluster_crds}")
    results = []
    for crd in custom_resource_definitions:
        if crd.kube_kind.singular not in cluster_crds:
            # TODO: kube_kind.singular seems to correspond to `crd.names.kind`
            # and not `crd.names.singular`
            log.warning(
                f"CRD {crd.kube_kind.singular} "
                f"not found in {cluster}",
            )
            continue
        config_dicts = load_all_configs(
            cluster=cluster,
            file_prefix=crd.file_prefix,
            soa_dir=soa_dir,
        )
        if not config_dicts:
            continue
        crs = list_custom_resources(
            kube_client=kube_client,
            kind=crd.kube_kind,
            version=crd.version,
            group=crd.group,
        )
        for cr in crs:
            service = config_dicts.get(cr.service)
            if service is not None:
                instance = service.get(cr.instance)
                if instance is not None:
                    continue
            result = False
            try:
                delete_custom_resource(
                    kube_client=kube_client,
                    name=cr.name,
                    namespace=cr.namespace,
                    plural=crd.kube_kind.plural,
                    version=crd.version,
                    group=crd.group,
                )
                result = True
            except Exception:
                log.exception('Error while deleting CR {cr.name}')
            results.append(result)
    return all(results) if results else True


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import yelp_meteorite

from paasta_tools import utils
from paasta_tools.smartstack_tools import retrieve_haproxy_csv


GUAGES = ['ereq', 'rate']
COUNTERS = ['hrsp_2xx', 'hrsp_3xx', 'hrsp_4xx', 'hrsp_5xx']


def parse_haproxy_backend_name(backend):
    # Currently a hack, the backend doesn't always match service.instance,
    # but at least the data will be reported.
    return backend.split('.')[0], backend.split('.')[1]


def report_metric_to_meteorite(backend, metric, value, paasta_cluster):
    try:
        paasta_service, paasta_instance = parse_haproxy_backend_name(backend)
    except IndexError:
        return

    meteorite_dims = {
        'paasta_service': paasta_service,
        'paasta_cluster': paasta_cluster,
        'paasta_instance': paasta_instance,
    }
    path = f'paasta.service.requests.{metric}'
    if metric in GUAGES:
        guage = yelp_meteorite.create_gauge(path, meteorite_dims)
        guage.set(value)
    elif metric in COUNTERS:
        counter = yelp_meteorite.create_counter(path, meteorite_dims)
        counter.count(value)
    else:
        raise ValueError(f"{metric} hasn't been configured as a guage or counter")
    print(f"Sent {path}: {value} to meteorite")


def report_all_metrics_to_meteorite(csv, paasta_cluster):
    for row in csv:
        if row['svname'] == 'BACKEND':
            for metric in GUAGES + COUNTERS:
                report_metric_to_meteorite(
                    backend=row['# pxname'],
                    metric=metric,
                    value=row[metric],
                    paasta_cluster=paasta_cluster,
                )


if __name__ == '__main__':
    system_paasta_config = utils.load_system_paasta_config()
    csv = retrieve_haproxy_csv(
        synapse_host=system_paasta_config.get_default_synapse_host(),
        synapse_port=system_paasta_config.get_synapse_port(),
        synapse_haproxy_url_format=system_paasta_config.get_synapse_haproxy_url_format(),
    )
    report_all_metrics_to_meteorite(
        csv=csv,
        paasta_cluster=system_paasta_config.get_local_run_config().get('default_cluster'),
    )



#!/usr/bin/env python
import argparse
import sys

from a_sync import block

from paasta_tools.autoscaling.autoscaling_cluster_lib import get_scaler
from paasta_tools.mesos.exceptions import MasterNotAvailableException
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.utils import load_system_paasta_config


def check_registration(threshold_percentage):
    try:
        mesos_state = block(get_mesos_master().state)
    except MasterNotAvailableException as e:
        print("Could not find Mesos Master: %s" % e.message)
        sys.exit(1)

    config = load_system_paasta_config()
    autoscaling_resources = config.get_cluster_autoscaling_resources()
    for resource in autoscaling_resources.values():
        print("Checking %s" % resource['id'])
        try:
            scaler = get_scaler(resource['type'])(
                resource=resource,
                pool_settings=None,
                config_folder=None,
                dry_run=True,
                utilization_error=0.0,
                max_increase=0.0,
                max_decrease=0.0,
            )
        except KeyError:
            print("Couldn't find a metric provider for resource of type: {}".format(resource['type']))
            continue
        if len(scaler.instances) == 0:
            print("No instances for this resource")
            continue
        elif scaler.is_new_autoscaling_resource():
            # See OPS-13784
            threshold = config.get_monitoring_config().get('check_registered_slave_threshold')
            print(
                f"Autoscaling resource was created within last {threshold}"
                " seconds and would probably fail this check",
            )
            continue
        else:
            slaves = scaler.get_aws_slaves(mesos_state)
            percent_registered = float(float(len(slaves)) / float(len(scaler.instances))) * 100
            if percent_registered < float(threshold_percentage):
                print("CRIT: Only found {}% of instances in {} registered in mesos. "
                      "Please check for puppet or AMI baking problems!".format(
                          percent_registered,
                          resource['id'],
                      ))
                return False
    print("OK: Found more than {}% of instances registered for all paasta resources in this "
          "superregion".format(threshold_percentage))
    return True


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t", "--threshold", help="percentage threshold for registered instances",
        default="75",
    )
    threshold = parser.parse_args().threshold
    if check_registration(threshold):
        sys.exit(0)
    sys.exit(2)


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Settings of the paasta-api server.
"""
import os
from typing import Optional

from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.marathon_tools import MarathonClients
from paasta_tools.utils import DEFAULT_SOA_DIR

soa_dir: str = os.environ.get("PAASTA_API_SOA_DIR", DEFAULT_SOA_DIR)

# The following `type: ignore` mypy hints are there because variables below de
# juro have `Optional[T]` type, but de facto are always initialized to a value
# of the corresponding type after the applicaton is started.
cluster: str = None  # type: ignore
marathon_clients: MarathonClients = None  # type: ignore
kubernetes_client: Optional[KubeClient] = None



import json
import os
from distutils.dir_util import copy_tree

from paasta_tools.cli.utils import pick_random_port


def main():
    print('-------------------------------------------------------')
    print('Please run export PAASTA_SYSTEM_CONFIG_DIR=etc_paasta_for_development to continue')
    print("Please set environment variable PAASTA_TEST_CLUSTER to the cluster you want to use.")
    print("This is necessary for tron jobs")
    print('-------------------------------------------------------')
    cluster = os.environ.get('PAASTA_TEST_CLUSTER', 'norcal-devc')
    config_path = 'etc_paasta_for_development'

    copy_tree('/etc/paasta', os.path.join(os.getcwd(), config_path))
    # Generate tron.json
    tron_config = {'tron': {'url': f'http://tron-{cluster}:8089'}}
    with open(config_path + '/tron.json', 'w') as f:
        json.dump(tron_config, f)
    # find unused port
    port = pick_random_port('paasta-dev-api')
    # Generate api endpoints
    api_endpoints = {'api_endpoints': {cluster: f'http://localhost:{port}'}}
    api_endpoints_path = os.path.join(os.getcwd(), config_path, 'api_endpoints.json')
    os.chmod(api_endpoints_path, 0o777)
    with open(api_endpoints_path, 'w') as f:
        json.dump(api_endpoints, f)

    # export config path
    os.environ['PAASTA_SYSTEM_CONFIG_DIR'] = config_path
    os.execl(
        '.tox/py36-linux/bin/python', '.tox/py36-linux/bin/python',
        '-m', 'paasta_tools.api.api', *['-D', '-c', cluster, str(port)],
    )


if __name__ == '__main__':
    main()



import asyncio
import functools
import time
import weakref
from collections import defaultdict
from typing import Any
from typing import AsyncIterable
from typing import Awaitable
from typing import Callable
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import TypeVar


T = TypeVar('T')


def async_ttl_cache(
    ttl: float = 300,
    cleanup_self: bool = False,
) -> Callable[
    [Callable[..., Awaitable[T]]],  # wrapped
    Callable[..., Awaitable[T]],  # inner
]:
    async def call_or_get_from_cache(cache, coro, args, kwargs):
        key = functools._make_key(args, kwargs, typed=False)
        try:
            future, last_update = cache[key]
            if ttl > 0 and time.time() - last_update > ttl:
                raise KeyError
        except KeyError:
            future = asyncio.ensure_future(coro)
            # set the timestamp to +infinity so that we always wait on the in-flight request.
            cache[key] = (future, float('Inf'))
        value = await future
        cache[key] = (future, time.time())
        return value

    if cleanup_self:
        cache: DefaultDict[Any, Dict] = defaultdict(dict)

        def on_delete(w):
            del cache[w]

        def outer(wrapped):
            @functools.wraps(wrapped)
            async def inner(self, *args, **kwargs):
                w = weakref.ref(self, on_delete)
                self_cache = cache[w]
                return await call_or_get_from_cache(
                    self_cache,
                    wrapped(self, *args, **kwargs),
                    args,
                    kwargs,
                )
            return inner
    else:
        cache2: Dict = {}  # Should be Dict[Any, T] but that doesn't work.

        def outer(wrapped):
            @functools.wraps(wrapped)
            async def inner(*args, **kwargs):
                return await call_or_get_from_cache(
                    cache2,
                    wrapped(*args, **kwargs),
                    args,
                    kwargs,
                )
            return inner
    return outer


async def aiter_to_list(
    aiter: AsyncIterable[T],
) -> List[T]:
    return [x async for x in aiter]


def async_timeout(
    seconds: int = 10,
) -> Callable[
    [Callable[..., Awaitable[T]]],  # wrapped
    Callable[..., Awaitable[T]],  # inner
]:
    def outer(wrapped):
        @functools.wraps(wrapped)
        async def inner(*args, **kwargs):
            return await asyncio.wait_for(wrapped(*args, **kwargs), timeout=seconds)
        return inner
    return outer



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to build a docker image."""
import os
import sys

from paasta_tools.cli.cmds.check import makefile_responds_to
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.utils import _log
from paasta_tools.utils import _log_audit
from paasta_tools.utils import _run
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_username
from paasta_tools.utils import paasta_print


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        'cook-image',
        description="Calls 'make cook-image' as part of the PaaSTA contract",
        help=(
            "'paasta cook-image' calls 'make cook-image' as part of the PaaSTA contract.\n\n"
            "The PaaSTA contract specifies that a service MUST respond to 'cook-image' and produce "
            "a docker image as a result. This command is often run as part of the normal build pipeline "
            "('paasta itest'), or via a 'paasta local-run --build'."
        ),
        epilog="This command assumes that the Makefile is in the current working directory.",
    )
    list_parser.add_argument(
        '-s', '--service',
        help=(
            'Build docker image for this service. Leading '
            '"services-", as included in a Jenkins job name, '
            'will be stripped.'
        ),
        required=True,
    )
    list_parser.add_argument(
        '-y', '--yelpsoa-config-root',
        dest='yelpsoa_config_root',
        help='A directory from which yelpsoa-configs should be read from',
        default=DEFAULT_SOA_DIR,
    )
    list_parser.set_defaults(command=paasta_cook_image)


def paasta_cook_image(args, service=None, soa_dir=None):
    """Build a docker image"""
    if not service:
        service = args.service
    if service.startswith('services-'):
        service = service.split('services-', 1)[1]
    if not soa_dir:
        soa_dir = args.yelpsoa_config_root
    validate_service_name(service, soa_dir)

    run_env = os.environ.copy()
    default_tag = 'paasta-cook-image-{}-{}'.format(service, get_username())
    tag = run_env.get('DOCKER_TAG', default_tag)
    run_env['DOCKER_TAG'] = tag

    if not makefile_responds_to('cook-image'):
        paasta_print(
            'ERROR: local-run now requires a cook-image target to be present in the Makefile. See'
            'http://paasta.readthedocs.io/en/latest/about/contract.html', file=sys.stderr,
        )
        return 1

    try:
        cmd = 'make cook-image'
        returncode, output = _run(
            cmd,
            env=run_env,
            log=True,
            component='build',
            service=service,
            loglevel='debug',
        )
        if returncode != 0:
            _log(
                service=service,
                line='ERROR: make cook-image failed for %s.' % service,
                component='build',
                level='event',
            )
        else:
            action_details = {
                'tag': tag,
            }
            _log_audit(
                action='cook-image',
                action_details=action_details,
                service=service,
            )
        return returncode

    except KeyboardInterrupt:
        paasta_print('\nProcess interrupted by the user. Cancelling.', file=sys.stderr)
        return 2
